/*
Title           :       DXARC.C

Purpose         :       Delete files extracted from specified archive.
                        Doesn't modify .ARC file in any way.  Reads .ARC
                        file and attempts to find and delete files whose
                        name, size, and timestamp match those in the .ARC
                        file.  Intended for easy cleanup of a directory 
                        containing files extracted from an .ARC file.

Author          :       Rowland Archer
                        CompuServe <70075,117>

Version         :       1.0   03/01/87

Usage           :       See usage() function.

Credits         :       J.A.Zucker's great library JAZLIB of C functions.
                        Vern Buerg's .ASM source for ARCV helped me figure
                        out ARC header format.  Thanks to both!

Build           :       You will need JAZLIB and Microsoft C 4.0 to use
                        this file with no changes.  JAZLIB is on many BBS's.

MAKE file       :       dxarc.obj:       dxarc.c
                            MSC dxarc /Zi /Od;

                        dxarc.exe:       dxarc.obj
                            LINK /CO dxarc,dxarc,dxarc/map,jzc jzscreen;

                        <set-up for CodeView debugging>
*/

#include <stdio.h>                       /* standard header file   */
#include <jaz.h>                         /* standard jaz header    */
#include <io.h>                          /* function declarations  */
#include <jzdirect.h>                    /* jaz directory defs     */
#include <direct.h>                      /* Msoft C directory defs */
#include <string.h>                      /* Msoft C string defs    */
#include <fcntl.h>                       /* Msoft C file ctrl defs */
#include <sys\types.h>                   /* Msoft C types          */
#include <sys\stat.h>                    /* Msoft C filestat block */

#define TRUE -1
#define FALSE 0
#define READ_ONLY 0
#define ARCMARK 26
#define ARCVER 9
#define SEEK_REL 1
#define FIND_NORM 0
#define FIND_ARC 32
#define MAXPATH 160
#define MAXFNAME 160
#define MAXSWITCH 2
#define DRIVELEN 2
#define SWITCHPERARG 5
#define MAXDRIVE 3
#define MAXARGS 10
#define AEMPTY 0
#define ADRIVE 1
#define APATH  2
#define AFNAME 4

struct archdr {                          /* Archive member header block */
    unsigned char mbrcode;
    char          mbrname[13];
    long          mbrsize;
    int           mbrdate;
    int           mbrtime;
    int           mbrcrc; 
    long          mbrlen;
} garchdr;

struct args {                            /* Struct to hold parsed cmd line */
    unsigned short int content;
    int     numsw;                       /* Number of switches in this entry */
    char    drive[MAXDRIVE];             /* E.g. "A:" */
    char    path[MAXPATH];               /* Pathname w/trailing '\'  */
    char    fname[MAXFNAME];             /* Filename */
    char    switches[SWITCHPERARG][MAXSWITCH];  /* Switches */
} gargs[MAXARGS];
/* content = bit map:   AEMPTY => nothing here
                        ADRIVE => drive spec
                        APATH  => path here
                        AFNAME => fname here   */

struct stat gstat;                       /* File status structure */

int  SCONFIRM = 0;                       /* Switch: confirm deletes */
int  SNODEL   = 0;                       /* Switch: no deletes, just report */
int  SMATCH   = 1;                       /* Switch: match creation time,
                                            creation date, and filesize
                                            before deleting file */

int  garch;                              /* ARC file handle      */
int  gnumdrives;                         /* Number of drives in system */
long garcoff;                            /* Offset into ARC file */
TDIR gdta;                               /* Used by JAZ find first/next */
char gcurdrive[MAXDRIVE];                /* Initial drive (0=A,1=B,etc) */
char gcurpath[MAXPATH];                  /* Initial path */
char garcdrive[MAXDRIVE];                /* .ARC file drive       */
char garcpath[MAXPATH];                  /* .ARC file path        */
char garcfpath[MAXDRIVE+MAXPATH+MAXFNAME]; /* .ARC file fullpath    */
char garcfname[MAXFNAME];                /* .ARC file name        */
char gclnpath[MAXPATH];                  /* Cleanup directory    */
char gclndrive[MAXDRIVE];                /* Cleanup drive        */
char gclnfpath[MAXDRIVE+MAXPATH];        /* Cleanup fullpath     */

int  getheader();                        /* Get archive header returns int */


main (argc,argv)
int argc;
char **argv;
{
    char wfname[MAXFNAME];               /* Temporary file name */
    char wpath[MAXPATH];                 /* Temporary pathname */
    char wmask[MAXFNAME];                /* Temporary file name  */
    char warcfpath[MAXDRIVE+MAXPATH+MAXFNAME];  /* Temp full path */
    char *woptptr;                       /* Pointer to option string */
    int  warcfound;                      /* Flag if archive name found */
    int  wcldfound;                      /* Flag if cleanup dir name found */
    int  wdrive;                         /* Drive number */
    int  wclndrive;                      /* Cleanup drive */
    int  i;                              /* Index var */
    int  werr;                           /* Error code */

    fprintf(stderr,"DXARC Ver. 1.0:  Delete Extracted .ARC files    3/01/87\n");
    fprintf(stderr,"Public Domain by Rowland Archer (CompuServe 70075,117)\n");
    if ((jzdosver() & 0xFF) < 2) {       /* Check DOS Version, quit if < 2.0 */
        printf("\nThis program requires DOS version 2.0 or greater to operate.\n");
        quit(1);
    }

    if (argc==1) {                       /* No args specified so give usage */
        usage();
        quit(1);
    }

    wdrive = jzgetdrv();                 /* Get initial logged drive */
    gcurdrive[0] = wdrive + 65;          /* 0=A, 1=B, etc              */
    gcurdrive[1] = ':';
    gcurdrive[2] = NULL;
    getcwd(gcurpath,MAXPATH);            /* Get initial logged path  */
    strcpy(gcurpath,&gcurpath[2]);       /* Strip drive from gcurpath */

    if (gcurpath[strlen(gcurpath)-1] != '\\')
        strcat(gcurpath,"\\");           /* Ensure final \ */

    strncpy(gclndrive,gcurdrive,MAXDRIVE);
    strncpy(gclnpath,gcurpath,MAXPATH);  /* Init cleanup dr/path to cur wkng */

    getargs(argc,argv);                  /* Get all args into arg structure */

    /* Scan args, looking for first filename:  use that as .ARC file */

    warcfound = FALSE;

    for (i=0;(!warcfound) && (i<argc-1); ++i) {
        if (gargs[i].content & (AFNAME | ADRIVE | APATH)) { /* fname template */
            if (*(gargs[i].fname)==NULL) { /* Check for null filename, illegal */
                fprintf(stderr,"\nYou must specify a .ARC file or template.\n");
                usage();
                quit(1);
            }

            getext(gargs[i].fname,".ARC");
            strncpy(garcfname,gargs[i].fname,MAXFNAME);

            if (gargs[i].content & ADRIVE)  /* Use arg's drive or cw drive */
                strncpy(garcdrive,gargs[i].drive,MAXDRIVE);
            else
                strncpy(garcdrive,gcurdrive,MAXDRIVE);

            if (gargs[i].content & APATH) {    /* Use arg's path or cwpath */
                strncpy(garcpath,gargs[i].path,MAXPATH);
                if (garcpath[strlen(garcpath)-1] != '\\')
                    if (strlen(garcpath) == MAXPATH) {
                        fprintf(stderr,"\nPathname too long: %s",
                          garcpath);
                        quit(1);
                    }
                    else
                        strcat(garcpath,"\\");   /* Ensure final '\' */
            }
            else if ((gargs[i].content & ADRIVE) == 0)  /* No path, no drive */
                strncpy(garcpath,gcurpath,MAXPATH);
            else                               /* Drive but no path spec'd */
                *garcpath = NULL;

            warcfound = TRUE;
        }
        if (gargs[i].numsw)   /* Handle switches if any found */
            setopt(i);
    }   /* for */

    if (!warcfound)  {        /* Then fell through argc w/out finding .ARC */
        fprintf(stderr,"\nYou must specify a .ARC file or template.\n");
        usage();
        quit(1);
    }

    /* Scan rest of args; next fname is cleanup dir; rest shld be switches */

    wcldfound = FALSE;
    for (;i < argc-1;++i) {
        if (wcldfound && (gargs[i].content & (AFNAME | ADRIVE | APATH))) {
            fprintf(stderr,"\nYou specified more than two pathnames or filenames.\n");
            usage();
            quit(1);
        }
        else if (!wcldfound && (gargs[i].content & (AFNAME | ADRIVE | APATH))) {

            if (gargs[i].content & ADRIVE) { /* Use arg's drive for cleanup */
                strncpy(gclndrive,gargs[i].drive,MAXDRIVE);
                *gclnpath = NULL;        /* Null out cleanup path, was cur wk */
            }

            if (gargs[i].content & APATH) { /* Use arg's path or cwpath */
                strncpy(gclnpath,gargs[i].path,MAXPATH);
                if (gclnpath[strlen(gclnpath)-1] != '\\')
                    if (strlen(gclnpath) == MAXPATH) {
                        fprintf(stderr,"\nPathname too long: %s",
                          gclnpath);
                        quit(1);
                    }
                    else
                        strcat(gclnpath,"\\");   /* Ensure final '\' */
            }

            if (gargs[i].content & AFNAME) { /* Can't specify a fname here */
                                             /* so must be last part of dir */
                strncat(gclnpath,gargs[i].fname,MAXPATH-strlen(gclnpath));
                if (gclnpath[strlen(gclnpath)-1] != '\\')
                    if (strlen(gclnpath) == MAXPATH) {
                        fprintf(stderr,"\nPathname too long: %s",
                          gclnpath);
                        quit(1);
                    }
                    else
                        strcat(gclnpath,"\\");           /* Ensure final \ */
            }
            wcldfound = TRUE;
        }
        if (gargs[i].numsw)                 /* Handle switches if any found */
            setopt(i);
    }   /* for */

    /*  Finally FINISHED parsing command line !!!   */
    /*  Start real work, looping on .ARC template */

    jzlogdrv(garcdrive[0]-'A');             /* log into .ARC drive */
    wdrive = jzgetdrv();                    /* see if we succeeded */

    if (wdrive != garcdrive[0]-'A') {
        fprintf(stderr,"\nDrive %c: does not exist",garcdrive[0]);
        quit(1);
    }

    /* Build up garcfpath, full path (not incl filename) of .ARC file(s) */
    
    *garcfpath = NULL;
    strcat (garcfpath,garcdrive);        /* Guaranteed to have an ARC drive */
    strcat (garcfpath,garcpath);         /* Pathname or null concat to fpath */

    if (*garcpath) {
        if ((garcpath[strlen(garcpath)-1] == '\\') &&
            (strlen(garcpath) > 1))
            werr = chdir(jzlftstr(garcpath,strlen(garcpath)-1));
        else
            werr = chdir(garcpath);             /* get into .ARC directory */
        if (werr) {
            fprintf(stderr,"\nCan't find directory: %c:%s\n",
                garcdrive[0],garcpath);
            quit(1);
        }
    }
    strcpy (warcfpath,garcfpath);        /* hold a copy of path name */
    werr = jzfndfst(garcfname,FIND_NORM | FIND_ARC,&gdta);
    if (!werr)
        do {
            strcat(garcfpath,gdta.name);
            garch = open(garcfpath,(O_RDONLY | O_BINARY));
            if (garch == -1) {
                fprintf(stderr,"\nCan't find directory %s",garcfpath);
                perror(" ");
            }
            else {
                wdrive = jzgetdrv();     /* save current drive */
                jzlogdrv(gclndrive[0]-'A');
                wclndrive = jzgetdrv();
                if (wclndrive != gclndrive[0]-'A') {
                    fprintf(stderr,"\nDrive %c: does not exist",gclndrive[0]);
                    quit(1);
                }
                jzlogdrv(wdrive);        /* restore drive */
                if (*gclnpath) {
                    *gclnfpath = NULL;
                    strcat(gclnfpath,gclndrive);
                    strcat(gclnfpath,gclnpath);
                    if (gclnfpath[strlen(gclnfpath)-1] == '\\' 
                                                && strlen(gclnpath) > 1)
                        gclnfpath[strlen(gclnfpath)-1] = NULL;
                    werr = stat(gclnfpath,&gstat);
                    if (werr || gstat.st_mode & S_IFREG) {
                        fprintf(stderr,"\nCan't find directory %s",gclnfpath);
                        quit(1);
                    }
                }
                fprintf(stderr,
                "\n---Using archive %s%s%s, working in directory %s%s\n",
                garcdrive,garcpath,gdta.name,gclndrive,gclnpath);
                procarc();               /* process archive */
                if (close(garch)) {      /* close archive file */
                    fprintf(stderr,"Error closing file %s",gdta.name);
                    perror(" ");
                }
            }
            werr = jzfndnxt(&gdta);
            strcpy(garcfpath,warcfpath); /* Reinitialize to drop fname */
        }  while (! werr);
    else {
        fprintf(stderr,"\nNo files matching %s%s%s found.\n",garcdrive,
            garcpath,garcfname);
        quit(1);
    }
    quit(0);
}



/*  Exit program and set errorlevel based on fexitcode */


quit(fexitcode)
int fexitcode;
{
    jzlogdrv(gcurdrive[0]-'A');          /* log back into original drive */
    if (*gcurpath)                       /* if there's a cur working path */
        if ((gcurpath[strlen(gcurpath)-1] == '\\') &&
            (strlen(gcurpath) > 1))
            chdir(jzlftstr(gcurpath,strlen(gcurpath)-1));
        else
            chdir(gcurpath);             /* get into original directory  */
    exit(fexitcode);
}



/*  Parse command line, load struct gargs with pathnames and options found */


getargs(argc,argv)                       /* Parse command line and load gargs */
int argc;                                /* w/upper-cased filenames, drives, */
char **argv;                             /* pathnames, and switches as-is */
{
    int  i;                              /* Index var */
    int  werr;                           /* Error code */
    char wfname[MAXFNAME];               /* Temporary file name */
    char *wfnameptr;                     /* Pointer to wfname */
    char wpath[MAXPATH];                 /* Temporary pathname */
    char *wpathptr;                      /* Pointer to wpath */
    char wmask[MAXFNAME];                /* Temporary file name  */
    char *woptptr;                       /* Pointer to opt string */

    for (i=0; i < argc-1; i++) {         /* For each argument on cmd line.. */
        gargs[i].numsw = 0;
        gargs[i].content = 0;
        strncpy(wmask,argv[i+1],MAXFNAME); /* Copy argument to check */
        if (*wmask == '/')               /* options, not a filename */
            getopts(wmask,i);            /* Parse opts and set numsw */
        else {                           /* Process as a drive/path/filename */
            jzprsfil(wmask,wpath,wfname);/* Get path and filename parts */
            wfnameptr = &wfname[0];
            if (*wfnameptr) {            /* If we have non-null wfname... */
                strupr(wfname);          /* Convert to upper case */
                woptptr = strpbrk(wfname,"/");  /* Options on filename? */
                if (woptptr != NULL) {
                    getopts(woptptr,i);  /* process the options */
                    *woptptr = NULL;     /* truncate options from arg */
                }
                strncpy(gargs[i].fname,wfname,MAXFNAME);
                gargs[i].content |= AFNAME;  /* Set bitmap => fname found */
            }

        /* Done with wfname.  Work on wpath and wdrive */

            wpathptr = &wpath[0];
            if (*wpathptr)               /* if there's a pathname arg..  */
                if (*(wpathptr+1) == ':') { /* and if there's a drive spec in it */
                    strncpy(gargs[i].drive,wpathptr,DRIVELEN);
                    strupr(gargs[i].drive);
                    gargs[i].content |= ADRIVE;
                    wpathptr += DRIVELEN;       /* step past drive spec */
                }
            if (*wpathptr) {                /* if still non-null string */
                strncpy(gargs[i].path,wpathptr,MAXPATH);
                strupr(gargs[i].path);
                gargs[i].content |= APATH;
            }
        }  /* else */
    }   /* for each arg in argv[i]... */
}   /* getargs() */



/*  Parse string starting with "/" and load options found into struct gargs */


getopts(foptptr,i)

char *foptptr;                           /* pointer to option char '/' */
int  i;                                  /* arg number in struct to set */
{
    char *wtoken;                        /* token = a switch */

    wtoken = strtok(foptptr,"/");        /* find switch delimited by '/' */
    while (wtoken != NULL) {
        if (gargs[i].numsw < SWITCHPERARG)
            strncpy(&gargs[i].switches[gargs[i].numsw++][0],wtoken,MAXSWITCH);
        wtoken = strtok(NULL,"/");
    }
}


/*  Set SWITCH variables based on options stored in struct gargs */


setopt(fargi)
int fargi;                               /* Global arg structure index */
{
    int i;

    /* step through gargs[fargi].numsw elements in gargs[fargi].switches */
    for (i=0; i < (gargs[fargi].numsw); ++i)
        switch((int)(*gargs[fargi].switches[i])) {
        case 'c':
        case 'C':                        /* Confirm deletes y/n */
            SCONFIRM = 1;
            break;
        case 'N':
        case 'n':                        /* No deletes, just report */
            SNODEL = 1;
            break;
        case 'F':
        case 'f':                        /* Force deletes, don't check */
            SMATCH = 0;                  /* match filesize, date, time */
            break;
        case '?':
            usage();
            quit(1);
            break;
        default:                         /* Unknown switch */
            fprintf(stderr,"\nThe option /%s is unknown.\n",
                gargs[fargi].switches[i]);
            usage();
            quit(1);
    }

}



/*  Add string at fext to ffname if ffname has no extension already */


getext(ffname,fext)
char *ffname;                            /* filename to get extension */
char *fext;                              /* .extension to add */
{
    char *wextptr;                       /* working pointer to extension */

    wextptr = strstr(ffname,fext);       /* look for extension in ffname */
    if (wextptr == NULL && strstr(ffname,".") == NULL)
        strcat(ffname,fext);             /* Add extension if needed */
}



/*  Print message explaining usage */


usage()
{
fprintf(stderr,"\nDXARC's purpose is to clean up a directory containing one or more files\n");
fprintf(stderr,"that have been extracted from an ARC file.  DXARC deletes files in the \n");
fprintf(stderr,"specified directory matching files in the specified archive.  It will \n");
fprintf(stderr,"only delete a file if its creation time, date, and size match those of \n");
fprintf(stderr,"the file in the archive, unless you specify the /F option.\n\n");
fprintf(stderr,"Usage: DXARC <arc_file_template>[.ARC] [<cleanup_directory>] [options]\n\n");
fprintf(stderr,"<arc_file_template>   Archive file name, may be simple filename, full\n");
fprintf(stderr,"                      pathname, or template with wildcards.  Default extension\n");
fprintf(stderr,"                      .ARC supplied if no extension given.\n");
fprintf(stderr,"<cleanup_directory>   Directory searched for files to delete.  Default is\n");
fprintf(stderr,"                      current working drive and directory.\n");
fprintf(stderr,"options         /C    Confirm by asking (yes/no) before deleting each file.\n");
fprintf(stderr,"   (prefix      /F    Force delete on matching filename even if there is \n");
fprintf(stderr,"   each option        a mismatch in creation date, time or file size.\n");
fprintf(stderr,"   with '/')    /N    No deletions at all, just report what would be deleted.\n");
}



/*  Process a single archive file  */


procarc()
{
    int  wc;                             /* temp to hold one character */
    int  wattr;                          /* file attributes */
    int *witime;                         /* file time stamp */
    int *widate;                         /* file date stamp */
    char wfullpath[MAXDRIVE+MAXPATH+MAXFNAME];  /* temporary full pathname */
    TDIR wdta;                           /* DTA for open file */

    while (getheader()) {                /* While there's another .ARC member */
        strncpy(wfullpath,gclndrive,MAXDRIVE);  /* Setup wfullpath with */
        strncat(wfullpath,gclnpath,MAXPATH);    /* cleanup drive, path, and */
        strncat(wfullpath,garchdr.mbrname,MAXFNAME);  /* .ARC member fname */

        fprintf(stderr,"%-30s : ",wfullpath);
        wattr = FIND_NORM | FIND_ARC;
        if (jzfndfst(wfullpath,wattr,&wdta)) {
            fprintf(stderr,"Not found\n");
            continue;
        }


        witime = (int *) &wdta.time;     /* Setup pointers to wfullpath's */
        widate = (int *) &wdta.date;     /* date and time stamp */

        if (SMATCH && (*widate != garchdr.mbrdate))  {
            fprintf(stderr,"Creation date differs, NOT deleted\n");
            continue;
        }

        if (SMATCH && (*witime != garchdr.mbrtime)) {
            fprintf(stderr,"Creation time differs, NOT deleted\n");
            continue;
        }

        if (garchdr.mbrcode != 1)  /* Old arc header didn't hold true fsize */
                                   /* so can't compare */
            if (SMATCH && (wdta.size != garchdr.mbrlen)) {
                fprintf(stderr,"File size differs, NOT deleted\n");
                continue;
            }

        if (SNODEL) {
            fprintf(stderr,"Found, NOT deleted\n");
            continue;
        }

        if (SCONFIRM) {
            fprintf(stderr,"Delete?  (Press Y or N): ");
            wc = toupper(getch());
            while (wc != 'Y' && wc != 'N') {
                jzbeep();
                wc = toupper(getch());
            }
            fprintf(stderr,"%c  ",wc);
            if (wc == 'Y')
                delfile(wfullpath);
            else
                fprintf(stderr,"\n");
        }
        else
            delfile(wfullpath);

    }   /* while(getheader()) */

}



/*  Delete the file ffullpath */


delfile(ffullpath)
char *ffullpath;
{
    int wresult;

    wresult = remove(ffullpath);
    if (wresult == 0)
        fprintf(stderr,"Deleted\n");
    else {
        fprintf(stderr,"NOT deleted");
        perror(" ");
    }
}



/*  Find next header in current archive file */


getheader()                              /* Find next header in .ARC file */
{
    char cbuf[1];                        /* buffer for file read */
    int  i;                              /* loop counter */
    int  wnum;                           /* number of bytes read */
    long wlnum;                          /* long return value for lseek */

    /* Find next header in .ARC file with handle garch */

    *cbuf = ARCMARK+1;         /* set to value <> ARCMARK so first test fails */
    i = 0;
    while ((*cbuf != ARCMARK) && (i++ <= 132)) {
        wnum = read(garch,cbuf,1);
        if (wnum==0 || wnum==-1) {           /* premature end of file */
            fprintf(stderr,"Invalid archive format in %s, premature end of file.\n",
            gdta.name);
            return(FALSE);
        }
    }

    /* Check if fell through counter w/out finding hdr */

    if (*cbuf != ARCMARK) {
        fprintf(stderr,
        "Invalid archive format in %s, archive header not found.\n",gdta.name);
        return(FALSE);
    }

    /* Found header, now check version code and load header struct if OK */

    wnum = read(garch,(char *) &garchdr.mbrcode,1);
    if (wnum==0 || wnum==-1) {           /* premature end of file */
        fprintf(stderr,"Invalid archive format in %s, premature end of file.\n",
        gdta.name);
        return(FALSE);
    }

    if (garchdr.mbrcode > ARCVER) { /* unknown archive format, skip this .ARC */
        fprintf(stderr,"Archive version unknown, skipping %s.\n", gdta.name);
        return(FALSE);
    }

    if (garchdr.mbrcode == 0) /* dummy member marks end of archive */
        return(FALSE);  /* no header found, but no errors either */

    wnum = read(garch,(char *) garchdr.mbrname,13);
    if (wnum!=13) {
        fprintf(stderr,"Invalid archive format in %s, premature end of file.\n",
        gdta.name);
        return(FALSE);
    }

    i = (garchdr.mbrcode==1 ? 10:14);    /* header size based on mbrcode */
    wnum = read(garch, (char *)&garchdr.mbrsize,i);
    if (wnum!=i) {
        fprintf(stderr,"Invalid archive format in %s, premature end of file.\n",
        gdta.name);
        return(FALSE);
    }
    if (i==10)
        garchdr.mbrlen = garchdr.mbrsize;
    wlnum = lseek(garch,(long) garchdr.mbrsize,SEEK_CUR);  /* seek past member */
    if (wlnum==-1L || eof(garch)) {
        fprintf(stderr,"Invalid archive format in %s, error attempting to position in file :",
        gdta.name);
        perror(" ");
        return(FALSE);
    }
    return(TRUE);                        /* Done, header filled */

}
